@@ -1,4 +1,4 @@
-/* $scrotwm: scrotwm.c,v 1.123 2009/02/01 04:03:40 marco Exp $ */
+/* $scrotwm: scrotwm.c,v 1.138 2009/02/03 22:19:40 marco Exp $ */
 /*
  * Copyright (c) 2009 Marco Peereboom <marco@peereboom.us>
  * Copyright (c) 2009 Ryan McBride <mcbride@countersiege.com>
@@ -48,9 +48,9 @@
  * DEALINGS IN THE SOFTWARE.
  */
 
-static const char	*cvstag = "$scrotwm: scrotwm.c,v 1.123 2009/02/01 04:03:40 marco Exp $";
+static const char	*cvstag = "$scrotwm: scrotwm.c,v 1.138 2009/02/03 22:19:40 marco Exp $";
 
-#define	SWM_VERSION	"0.7"
+#define	SWM_VERSION	"0.8"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -128,9 +128,9 @@ u_int32_t		swm_debug = 0
 #define BUTTONMASK		(ButtonPressMask|ButtonReleaseMask)
 #define MOUSEMASK		(BUTTONMASK|PointerMotionMask)
 #define SWM_PROPLEN		(16)
-#define X(r)			(r)->g.x	
+#define X(r)			(r)->g.x
 #define Y(r)			(r)->g.y
-#define WIDTH(r)		(r)->g.w	
+#define WIDTH(r)		(r)->g.w
 #define HEIGHT(r)		(r)->g.h
 
 #ifndef SWM_LIB
@@ -143,6 +143,7 @@ int			(*xerrorxlib)(Display *, XErrorEvent *);
 int			other_wm;
 int			running = 1;
 int			ss_enabled = 0;
+int			xrandr_support;
 int			xrandr_eventbase;
 int			ignore_enter = 0;
 unsigned int		numlockmask = 0;
@@ -181,6 +182,8 @@ char			*bar_fonts[] = {
 /* terminal + args */
 char			*spawn_term[] = { "xterm", NULL };
 char			*spawn_screenshot[] = { "screenshot.sh", NULL, NULL };
+char			*spawn_lock[] = { "xlock", NULL };
+char			*spawn_initscr[] = { "initscreen.sh", NULL };
 char			*spawn_menu[] = { "dmenu_run", "-fn", NULL, "-nb", NULL,
 			    "-nf", NULL, "-sb", NULL, "-sf", NULL, NULL };
 
@@ -205,9 +208,9 @@ struct workspace;
 struct swm_region {
 	TAILQ_ENTRY(swm_region)	entry;
 	struct swm_geometry	g;
-	Window			bar_window;
 	struct workspace 	*ws;	/* current workspace on this region */
 	struct swm_screen	*s;	/* screen idx */
+	Window			bar_window;
 }; 
 TAILQ_HEAD(swm_region_list, swm_region);
 
@@ -219,6 +222,7 @@ struct ws_win {
 	int			floating;
 	int			transient;
 	int			manual;
+	unsigned long		quirks;
 	struct workspace	*ws;	/* always valid */
 	struct swm_screen	*s;	/* always valid, never changes */
 	XWindowAttributes	wa;
@@ -264,8 +268,10 @@ struct workspace {
 	struct {
 				int horizontal_msize;
 				int horizontal_mwin;
+				int horizontal_stacks;
 				int vertical_msize;
 				int vertical_mwin;
+				int vertical_stacks;
 	} l_state;
 };
 
@@ -277,8 +283,8 @@ enum	{ SWM_S_COLOR_BAR, SWM_S_COLOR_BAR_BORDER, SWM_S_COLOR_BAR_FONT,
 struct swm_screen {
 	int			idx;		/* screen index */
 	struct swm_region_list	rl;	/* list of regions on this screen */
+	struct swm_region_list	orl;	/* list of old regions */
 	Window			root;
-	int			xrandr_support;
 	struct workspace	ws[SWM_WS_MAX];
 
 	/* colors */
@@ -311,6 +317,10 @@ union arg {
 #define SWM_ARG_ID_CYCLEWS_DOWN	(13)
 #define SWM_ARG_ID_CYCLESC_UP	(14)
 #define SWM_ARG_ID_CYCLESC_DOWN	(15)
+#define SWM_ARG_ID_COLINC	(16)
+#define SWM_ARG_ID_COLDEC	(17)
+#define SWM_ARG_ID_ROWINC	(16)
+#define SWM_ARG_ID_ROWDEL	(17)
 #define SWM_ARG_ID_SS_ALL	(0)
 #define SWM_ARG_ID_SS_WINDOW	(1)
 #define SWM_ARG_ID_DONTCENTER	(0)
@@ -323,11 +333,16 @@ struct quirk {
 	char			*class;
 	char			*name;
 	unsigned long		quirk;
-#define SWM_Q_FLOAT		(1<<0)
+#define SWM_Q_FLOAT		(1<<0)	/* float this window */
+#define SWM_Q_TRANSSZ		(1<<1)	/* transiend window size too small */
+#define SWM_Q_ANYWHERE		(1<<2)	/* don't position this window */
 } quirks[] = {
 	{ "MPlayer",		"xv",		SWM_Q_FLOAT },
 	{ "OpenOffice.org 2.4",	"VCLSalFrame",	SWM_Q_FLOAT },
-	{ NULL,		NULL,		0},
+	{ "OpenOffice.org 3.0",	"VCLSalFrame",	SWM_Q_FLOAT },
+	{ "Firefox-bin",	"firefox-bin",	SWM_Q_TRANSSZ},
+	{ "Gimp",		"gimp",		SWM_Q_FLOAT | SWM_Q_ANYWHERE},
+	{ NULL,			NULL,		0},
 };
 
 /* events */
@@ -403,6 +418,35 @@ setscreencolor(char *val, int i, int c)
 		    i, ScreenCount(display));
 }
 
+void		new_region(struct swm_screen *, int, int, int, int);
+
+void
+custom_region(char *val)
+{
+	unsigned int			sidx, x, y, w, h;
+
+	if (sscanf(val, "screen[%u]:%ux%u+%u+%u", &sidx, &w, &h, &x, &y) != 5)
+		errx(1, "invalid custom region, "
+		    "should be 'screen[<n>]:<n>x<n>+<n>+<n>\n");
+	if (sidx < 1 || sidx > ScreenCount(display))
+		errx(1, "invalid screen index: %d out of bounds (maximum %d)\n",
+		    sidx, ScreenCount(display));
+	sidx--;
+
+	if (w < 1 || h < 1)
+		errx(1, "region %ux%u+%u+%u too small\n", w, h, x, y);
+
+	if (x  < 0 || x > DisplayWidth(display, sidx) ||
+	    y < 0 || y > DisplayHeight(display, sidx) ||
+	    w + x > DisplayWidth(display, sidx) ||
+	    h + y > DisplayHeight(display, sidx))
+		errx(1, "region %ux%u+%u+%u not within screen boundaries "
+		    "(%ux%u)\n", w, h, x, y,
+		    DisplayWidth(display, sidx), DisplayHeight(display, sidx));
+	    
+	new_region(&screens[sidx], x, y, w, h);
+}
+
 int
 varmatch(char *var, char *name, int *index)
 {
@@ -509,6 +553,13 @@ conf_load(char *filename)
 				goto bad;
 			break;
 
+		case 'r':
+			if (!strncmp(var, "region", strlen("region")))
+				custom_region(val);
+			else
+				goto bad;
+			break;
+
 		case 's':
 			if (!strncmp(var, "spawn_term", strlen("spawn_term")))
 				asprintf(&spawn_term[0], "%s", val);
@@ -822,7 +873,7 @@ restart(struct swm_region *r, union arg *args)
 struct swm_region *
 root_to_region(Window root)
 {
-	struct swm_region	*r;
+	struct swm_region	*r = NULL;
 	Window			rr, cr;
 	int			i, x, y, wx, wy;
 	unsigned int		mask;
@@ -832,20 +883,17 @@ root_to_region(Window root)
 			break;
 
 	if (XQueryPointer(display, screens[i].root, 
-	    &rr, &cr, &x, &y, &wx, &wy, &mask) == False) {
-		/* if we can't query the pointer, grab the first region */
-		r = TAILQ_FIRST(&screens[i].rl);
-	} else {
-		/* otherwise, choose a region based on pointer location */
-		TAILQ_FOREACH(r, &screens[i].rl, entry) {
+	    &rr, &cr, &x, &y, &wx, &wy, &mask) != False) {
+		/* choose a region based on pointer location */
+		TAILQ_FOREACH(r, &screens[i].rl, entry)
 			if (x >= X(r) && x <= X(r) + WIDTH(r) &&
 			    y >= Y(r) && y <= Y(r) + HEIGHT(r))
 				break;
-		}
-
-		if (r == NULL)
-			r = TAILQ_FIRST(&screens[i].rl);
 	}
+
+	if (r == NULL)
+		r = TAILQ_FIRST(&screens[i].rl);
+
 	return (r);
 }
 
@@ -974,7 +1022,7 @@ switchws(struct swm_region *r, union arg *args)
 	old_ws = this_r->ws;
 	new_ws = &this_r->s->ws[wsid];
 
-	DNPRINTF(SWM_D_WS, "switchws screen %d region %dx%d+%d+%d: "
+	DNPRINTF(SWM_D_WS, "switchws screen[%d]:%dx%d+%d+%d: "
 	    "%d -> %d\n", r->s->idx, WIDTH(r), HEIGHT(r), X(r), Y(r),
 	    old_ws->idx, wsid);
 
@@ -985,13 +1033,14 @@ switchws(struct swm_region *r, union arg *args)
 	if (!other_r) {
 		/* if the other workspace is hidden, switch windows */
 		/* map new window first to prevent ugly blinking */
+		old_ws->r = NULL;
+		old_ws->restack = 1;
+
 		TAILQ_FOREACH(win, &new_ws->winlist, entry)
 			XMapRaised(display, win->id);
 
 		TAILQ_FOREACH(win, &old_ws->winlist, entry)
 			XUnmapWindow(display, win->id);
-		old_ws->r = NULL;
-		old_ws->restack = 1;
 	} else {
 		other_r->ws = old_ws;
 		old_ws->r = other_r;
@@ -1014,7 +1063,7 @@ cyclews(struct swm_region *r, union arg *args)
 	struct swm_screen	*s = r->s;
 
 	DNPRINTF(SWM_D_WS, "cyclews id %d "
-	    "in screen %d region %dx%d+%d+%d ws %d\n", args->id,
+	    "in screen[%d]:%dx%d+%d+%d ws %d\n", args->id,
 	    r->s->idx, WIDTH(r), HEIGHT(r), X(r), Y(r), r->ws->idx);
 
 	a.id = r->ws->idx;
@@ -1233,7 +1282,7 @@ stack_floater(struct ws_win *win, struct swm_region *r)
 	bzero(&wc, sizeof wc);
 	mask = CWX | CWY | CWBorderWidth | CWWidth | CWHeight;
 	wc.border_width = 1;
-	if (win->transient) {
+	if (win->transient && (win->quirks & SWM_Q_TRANSSZ)) {
 		win->g.w = (double)WIDTH(r) * dialog_ratio;
 		win->g.h = (double)HEIGHT(r) * dialog_ratio;
 	}
@@ -1264,7 +1313,7 @@ stack_master(struct workspace *ws, struct swm_geometry *g, int rot, int flip)
 	XWindowChanges		wc;
 	struct swm_geometry	win_g, r_g = *g;
 	struct ws_win		*win, *winfocus;
-	int			i, j, w_inc, h_inc, w_base, h_base;
+	int			i, j, s, w_inc, h_inc, w_base, h_base, stacks; 
 	int			hrh, extra, h_slice, last_h = 0;
 	int			split, colno, winno, mwin, msize, mscale;
 	int			remain, missing, v_slice;;
@@ -1286,22 +1335,27 @@ stack_master(struct workspace *ws, struct swm_geometry *g, int rot, int flip)
 		w_base = win->sh.base_width;
 		mwin = ws->l_state.horizontal_mwin;
 		mscale = ws->l_state.horizontal_msize;
+		stacks = ws->l_state.horizontal_stacks;
 		SWAPXY(&r_g);
 	} else {
 		w_inc = win->sh.height_inc;
 		w_base = win->sh.base_height;
 		mwin = ws->l_state.vertical_mwin;
 		mscale = ws->l_state.vertical_msize;
+		stacks = ws->l_state.vertical_stacks;
 	}
 	win_g = r_g;
 
+	if (stacks > winno - mwin)
+		stacks = winno - mwin;
+
 	h_slice = r_g.h / SWM_H_SLICE;
 	if (mwin && winno > mwin) {
 		v_slice = r_g.w / SWM_V_SLICE;
 
 		split = mwin;
 		colno = split;
-		msize = v_slice * mscale;
+		win_g.w = v_slice * mscale;
 
 		if (w_inc > 1 && w_inc < v_slice) {
 			/* adjust for window's requested size increment */
@@ -1317,32 +1371,44 @@ stack_master(struct workspace *ws, struct swm_geometry *g, int rot, int flip)
 			}
 		}
 
-		win_g.w = msize;
+		msize = win_g.w;
 		if (flip) 
 			win_g.x += r_g.w - msize;
 	} else {
-		colno = winno;
-		split = 0;
+		if (stacks > 1) {
+			colno = split = (winno - mwin) / stacks;
+		} else {
+			split = 0;
+			colno = winno;
+		}
 	}
 	hrh = r_g.h / colno;
 	extra = r_g.h - (colno * hrh);
 	win_g.h = hrh - 2;
 
 	/*  stack all the tiled windows */
-	i = j = 0;
+	i = j = 0, s = stacks;
 	TAILQ_FOREACH(win, &ws->winlist, entry) {
 		if (win->transient != 0 || win->floating != 0)
 			continue;
 
 		if (split && i == split) {
-			colno = winno - split;
+			colno = (winno - mwin) / stacks;
+			if (s <= (winno - mwin) % stacks)
+				colno++;
+			split = split + colno;
 			hrh = (r_g.h / colno);
 			extra = r_g.h - (colno * hrh);
 			if (flip)
 				win_g.x = r_g.x;
 			else
-				win_g.x += msize + 2;
-			win_g.w = r_g.w - (msize + 2);
+				win_g.x += win_g.w + 2;
+			win_g.w = (((r_g.w - (msize + 2)) -
+			    ((stacks - 1) * 2)) / stacks);
+			if (s == 1)
+				win_g.w += (((r_g.w - (msize + 2)) -
+				    ((stacks - 1) * 2)) % stacks);
+			s--;
 			j = 0;
 		}
 		win_g.h = hrh - 2;
@@ -1350,7 +1416,7 @@ stack_master(struct workspace *ws, struct swm_geometry *g, int rot, int flip)
 			h_inc = win->sh.width_inc;
 			h_base = win->sh.base_width;
 		} else {
-		   	h_inc =	win->sh.height_inc;	
+		   	h_inc =	win->sh.height_inc;
 			h_base = win->sh.base_height;
 		}
 		if (j == colno - 1) {
@@ -1390,9 +1456,6 @@ stack_master(struct workspace *ws, struct swm_geometry *g, int rot, int flip)
 		mask = CWX | CWY | CWWidth | CWHeight | CWBorderWidth;
 		XConfigureWindow(display, win->id, mask, &wc);
 		XMapRaised(display, win->id);
-		/*
-		fprintf(stderr, "vertical_stack: win %d x %d y %d w %d h %d bw %d\n", win->id, win->g.x, win->g.y, win->g.w , win->g.h, wc.border_width);
-		*/
 
 		last_h = win_g.h;
 		i++;
@@ -1422,6 +1485,7 @@ vertical_config(struct workspace *ws, int id)
 	case SWM_ARG_ID_STACKINIT:
 		ws->l_state.vertical_msize = SWM_V_SLICE / 2;
 		ws->l_state.vertical_mwin = 1;
+		ws->l_state.vertical_stacks = 1;
 		break;
 	case SWM_ARG_ID_MASTERSHRINK:
 		if (ws->l_state.vertical_msize > 1)
@@ -1437,6 +1501,12 @@ vertical_config(struct workspace *ws, int id)
 	case SWM_ARG_ID_MASTERDEL:
 		if (ws->l_state.vertical_mwin > 0)
 			ws->l_state.vertical_mwin--;
+	case SWM_ARG_ID_COLINC:
+		ws->l_state.vertical_stacks++;
+		break;
+	case SWM_ARG_ID_COLDEC:
+		if (ws->l_state.vertical_stacks > 1)
+			ws->l_state.vertical_stacks--;
 		break;
 	default:
 		return;
@@ -1461,6 +1531,7 @@ horizontal_config(struct workspace *ws, int id)
 	case SWM_ARG_ID_STACKINIT:
 		ws->l_state.horizontal_mwin = 1;
 		ws->l_state.horizontal_msize = SWM_H_SLICE / 2;
+		ws->l_state.horizontal_stacks = 1;
 		break;
 	case SWM_ARG_ID_MASTERSHRINK:
 		if (ws->l_state.horizontal_msize > 1)
@@ -1477,6 +1548,12 @@ horizontal_config(struct workspace *ws, int id)
 		if (ws->l_state.horizontal_mwin > 0)
 			ws->l_state.horizontal_mwin--;
 		break;
+	case SWM_ARG_ID_COLINC:
+		ws->l_state.horizontal_stacks++;
+		break;
+	case SWM_ARG_ID_COLDEC:
+		if (ws->l_state.horizontal_stacks > 1)
+			ws->l_state.horizontal_stacks--;
 	default:
 		return;
 	}
@@ -1645,6 +1722,8 @@ struct key {
 	{ MODKEY,		XK_l,		stack_config,	{.id = SWM_ARG_ID_MASTERGROW} },
 	{ MODKEY,		XK_comma,	stack_config,	{.id = SWM_ARG_ID_MASTERADD} },
 	{ MODKEY,		XK_period,	stack_config,	{.id = SWM_ARG_ID_MASTERDEL} },
+	{ MODKEY | ShiftMask,	XK_comma,	stack_config,	{.id = SWM_ARG_ID_COLINC} },
+	{ MODKEY | ShiftMask,	XK_period,	stack_config,	{.id = SWM_ARG_ID_COLDEC} },
 	{ MODKEY,		XK_Return,	swapwin,	{.id = SWM_ARG_ID_SWAPMAIN} },
 	{ MODKEY,		XK_j,		focus,		{.id = SWM_ARG_ID_FOCUSNEXT} },
 	{ MODKEY,		XK_k,		focus,		{.id = SWM_ARG_ID_FOCUSPREV} },
@@ -1687,6 +1766,8 @@ struct key {
 	{ MODKEY | ShiftMask,	XK_s,		screenshot,	{.id = SWM_ARG_ID_SS_WINDOW} },
 	{ MODKEY,		XK_t,		floating_toggle,{0} },
 	{ MODKEY | ShiftMask,	XK_v,		version,	{0} },
+	{ MODKEY | ShiftMask,	XK_Delete,	spawn,		{.argv = spawn_lock} },
+	{ MODKEY | ShiftMask,	XK_i,		spawn,		{.argv = spawn_initscr} },
 };
 
 void
@@ -1721,7 +1802,7 @@ resize(struct ws_win *win, union arg *args)
 	XEvent			ev;
 	Time			time = 0;
 
-	DNPRINTF(SWM_D_MOUSE, "resize: win %d floating %d trans %d\n",
+	DNPRINTF(SWM_D_MOUSE, "resize: win %lu floating %d trans %d\n",
 	    win->id, win->floating, win->transient);
 
 	if (!(win->transient != 0 || win->floating != 0))
@@ -1796,7 +1877,7 @@ move(struct ws_win *win, union arg *args)
 	Time			time = 0;
 	int			restack = 0;
 
-	DNPRINTF(SWM_D_MOUSE, "move: win %d floating %d trans %d\n",
+	DNPRINTF(SWM_D_MOUSE, "move: win %lu floating %d trans %d\n",
 	    win->id, win->floating, win->transient);
 
 	if (win->floating == 0) {
@@ -1999,6 +2080,7 @@ manage_window(Window id)
 	struct swm_region	*r;
 	long			mask;
 	const char		*errstr;
+	XWindowChanges		wc;
 
 	if ((win = find_window(id)) != NULL)
 			return (win);	/* already being managed */
@@ -2019,7 +2101,7 @@ manage_window(Window id)
 		DNPRINTF(SWM_D_MISC, "manage_window: win %u transient %u\n",
 		    (unsigned)win->id, win->transient);
 	}
-	
+
 	/*
 	 * Figure out where to put the window. If it was previously assigned to
 	 * a workspace (either by spawn() or manually moving), and isn't
@@ -2058,10 +2140,6 @@ manage_window(Window id)
 	}
 	XFree(prop);
 
-	/*
-	fprintf(stderr, "manage window: %d x %d y %d w %d h %d\n", win->id, win->g.x, win->g.y, win->g.w, win->g.h);
-	*/
-
 	if (XGetClassHint(display, win->id, &win->ch)) {
 		DNPRINTF(SWM_D_CLASS, "class: %s name: %s\n",
 		    win->ch.res_class, win->ch.res_name);
@@ -2073,21 +2151,79 @@ manage_window(Window id)
 				    win->ch.res_class, win->ch.res_name);
 				if (quirks[i].quirk & SWM_Q_FLOAT)
 					win->floating = 1;
+				win->quirks = quirks[i].quirk;
 			}
 		}
 	}
 
+	/* alter window position if quirky */
+	if (win->quirks & SWM_Q_ANYWHERE) {
+		win->manual = 1; /* don't center the quirky windows */
+		bzero(&wc, sizeof wc);
+		mask = 0;
+		if (win->g.y < bar_height) {
+			win->g.y = wc.y = bar_height;
+			mask |= CWY;
+		}
+		if (win->g.w + win->g.x > WIDTH(r)) {
+			win->g.x = wc.x = WIDTH(win->ws->r) - win->g.w - 2;
+			mask |= CWX;
+		}
+		wc.border_width = 1;
+		mask |= CWBorderWidth;
+		XConfigureWindow(display, win->id, mask, &wc);
+	}
+
 	XSelectInput(display, id, EnterWindowMask | FocusChangeMask |
 	    PropertyChangeMask | StructureNotifyMask);
 
 	set_win_state(win, NormalState);
 
+	/* floaters need to be mapped if they are in the current workspace */
+	if (win->floating && (ws->idx == r->ws->idx))
+		XMapRaised(display, win->id);
+
 	/* make new win focused */
 	focus_win(win);
 
 	return (win);
 }
 
+void
+unmanage_window(struct ws_win *win)
+{
+	struct workspace	*ws;
+
+	if (win == NULL)
+		return;
+
+	DNPRINTF(SWM_D_MISC, "unmanage_window:  %lu\n", win->id);
+
+	/* don't unmanage if we are switching workspaces */
+	ws = win->ws;
+	if (ws->restack)
+		return;
+
+	/* find a window to focus */
+	if (ws->focus == win)
+		ws->focus = TAILQ_PREV(win, ws_win_list, entry);
+	if (ws->focus == NULL)
+		ws->focus = TAILQ_FIRST(&ws->winlist);
+	if (ws->focus == NULL || ws->focus == win) {
+		ws->focus = NULL;
+		unfocus_all();
+	} else
+		focus_win(ws->focus);
+
+	TAILQ_REMOVE(&win->ws->winlist, win, entry);
+	set_win_state(win, WithdrawnState);
+	if (win->ch.res_class)
+		XFree(win->ch.res_class);
+	if (win->ch.res_name)
+		XFree(win->ch.res_name);
+	free(win);
+}
+
 void
 configurerequest(XEvent *e)
 {
@@ -2102,10 +2238,6 @@ configurerequest(XEvent *e)
 	if (new) {
 		DNPRINTF(SWM_D_EVENT, "configurerequest: new window: %lu\n",
 		    ev->window);
-		/*
-		fprintf(stderr, "configurerequest: new window: %lu x %d y %d w %d h %d bw %d s %d sm %d\n",
-		    ev->window, ev->x, ev->y, ev->width, ev->height, ev->border_width, ev->above, ev->detail);
-		*/
 		bzero(&wc, sizeof wc);
 		wc.x = ev->x;
 		wc.y = ev->y;
@@ -2116,10 +2248,6 @@ configurerequest(XEvent *e)
 		wc.stack_mode = ev->detail;
 		XConfigureWindow(display, ev->window, ev->value_mask, &wc);
 	} else {
-		/*
-		fprintf(stderr, "configurerequest: change window: %lu\n",
-		    ev->window);
-		*/
 		DNPRINTF(SWM_D_EVENT, "configurerequest: change window: %lu\n",
 		    ev->window);
 		if (win->floating) {
@@ -2145,8 +2273,8 @@ configurerequest(XEvent *e)
 					ev->value_mask |= CWY | CWHeight;
 				}
 			}
-			if ((ev->value_mask & (CWX|CWY)) &&
-			    !(ev->value_mask & (CWWidth|CWHeight)))
+			if ((ev->value_mask & (CWX | CWY)) &&
+			    !(ev->value_mask & (CWWidth | CWHeight)))
 				config_win(win);
 			XMoveResizeWindow(display, win->id,
 			    win->g.x, win->g.y, win->g.w, win->g.h);
@@ -2167,29 +2295,11 @@ destroynotify(XEvent *e)
 {
 	struct ws_win		*win;
 	XDestroyWindowEvent	*ev = &e->xdestroywindow;
-	struct workspace	*ws;
 
 	DNPRINTF(SWM_D_EVENT, "destroynotify: window %lu\n", ev->window);
 
 	if ((win = find_window(ev->window)) != NULL) {
-		ws = win->ws;
-		/* find a window to focus */
-		if (ws->focus == win)
-			ws->focus = TAILQ_PREV(win, ws_win_list, entry);
-		if (ws->focus == NULL)
-			ws->focus = TAILQ_FIRST(&ws->winlist);
-		if (ws->focus == NULL || ws->focus == win) {
-			ws->focus = NULL;
-			unfocus_all();
-		} else
-			focus_win(ws->focus);
-		TAILQ_REMOVE(&ws->winlist, win, entry);
-		set_win_state(win, WithdrawnState);
-		if (win->ch.res_class)
-			XFree(win->ch.res_class);
-		if (win->ch.res_name)
-			XFree(win->ch.res_name);
-		free(win);
+		unmanage_window(win);
 		stack();
 	}
 }
@@ -2244,6 +2354,7 @@ maprequest(XEvent *e)
 	if (wa.override_redirect)
 		return;
 	manage_window(e->xmaprequest.window);
+
 	stack();
 }
 
@@ -2285,7 +2396,14 @@ propertynotify(XEvent *e)
 void
 unmapnotify(XEvent *e)
 {
+	XDestroyWindowEvent	*ev = &e->xdestroywindow;
+	struct ws_win		*win;
+
 	DNPRINTF(SWM_D_EVENT, "unmapnotify: window: %lu\n", e->xunmap.window);
+
+	if ((win = find_window(ev->window)) != NULL)
+		if (win->transient)
+			unmanage_window(win);
 }
 
 void
@@ -2356,16 +2474,63 @@ getstate(Window w)
 }
 
 void
-new_region(struct swm_screen *s, struct workspace *ws,
-    int x, int y, int w, int h)
+new_region(struct swm_screen *s, int x, int y, int w, int h)
 {
-	struct swm_region	*r;
+	struct swm_region	*r, *n;
+	struct workspace	*ws = NULL;
+	int			i;
 
-	DNPRINTF(SWM_D_MISC, "new region on screen %d: %dx%d (%d, %d)\n",
-	     s->idx, x, y, w, h);
+	DNPRINTF(SWM_D_MISC, "new region: screen[%d]:%dx%d+%d+%d\n",
+	     s->idx, w, h, x, y);
 
-	if ((r = calloc(1, sizeof(struct swm_region))) == NULL)
-		errx(1, "calloc: failed to allocate memory for screen");
+	/* remove any conflicting regions */
+	n = TAILQ_FIRST(&s->rl);
+	while (n) {
+		r = n;
+		n = TAILQ_NEXT(r, entry);
+		if (X(r) < (x + w) &&
+		    (X(r) + WIDTH(r)) > x &&
+		    Y(r) < (y + h) &&
+		    (Y(r) + HEIGHT(r)) > y) {
+			XDestroyWindow(display, r->bar_window);
+			TAILQ_REMOVE(&s->rl, r, entry);
+			TAILQ_INSERT_TAIL(&s->orl, r, entry);
+		}
+	}
+
+	/* search old regions for one to reuse */
+
+	/* size + location match */
+	TAILQ_FOREACH(r, &s->orl, entry)
+		if (X(r) == x && Y(r) == y &&
+		    HEIGHT(r) == h && WIDTH(r) == w)
+			break;
+
+	/* size match */
+	TAILQ_FOREACH(r, &s->orl, entry)
+		if (HEIGHT(r) == h && WIDTH(r) == w)
+			break;
+
+	if (r != NULL) {
+		TAILQ_REMOVE(&s->orl, r, entry);
+		/* try to use old region's workspace */
+		if (r->ws->r == NULL)
+			ws = r->ws;
+	} else
+		if ((r = calloc(1, sizeof(struct swm_region))) == NULL)
+			errx(1, "calloc: failed to allocate memory for screen");
+
+	/* if we don't have a workspace already, find one */
+	if (ws == NULL) {
+		for (i = 0; i < SWM_WS_MAX; i++)
+			if (s->ws[i].r == NULL) {
+				ws = &s->ws[i];
+				break;
+			}
+	}
+
+	if (ws == NULL)
+		errx(1, "no free workspaces\n");
 
 	X(r) = x;
 	Y(r) = y;
@@ -2379,20 +2544,100 @@ new_region(struct swm_screen *s, struct workspace *ws,
 }
 
 void
-setup_screens(void)
+scan_xrandr(int i)
 {
 #ifdef SWM_XRR_HAS_CRTC
 	XRRCrtcInfo		*ci;
 	XRRScreenResources	*sr;
 	int			c;
+	int			ncrtc = 0;
 #endif /* SWM_XRR_HAS_CRTC */
+	struct swm_region	*r;
+
+
+	if (i >= ScreenCount(display))
+		errx(1, "invalid screen");
+
+	/* remove any old regions */
+	while ((r = TAILQ_FIRST(&screens[i].rl)) != NULL) {
+		r->ws->r = NULL;
+		XDestroyWindow(display, r->bar_window);
+		TAILQ_REMOVE(&screens[i].rl, r, entry);
+		TAILQ_INSERT_TAIL(&screens[i].orl, r, entry);
+	}
+
+	/* map virtual screens onto physical screens */
+#ifdef SWM_XRR_HAS_CRTC
+	if (xrandr_support) {
+		sr = XRRGetScreenResources(display, screens[i].root);
+		if (sr == NULL)
+			new_region(&screens[i], 0, 0,
+			    DisplayWidth(display, i),
+			    DisplayHeight(display, i)); 
+		else 
+			ncrtc = sr->ncrtc;
+
+		for (c = 0, ci = NULL; c < ncrtc; c++) {
+			ci = XRRGetCrtcInfo(display, sr, sr->crtcs[c]);
+			if (ci->noutput == 0)
+				continue;
+
+			if (ci != NULL && ci->mode == None)
+				new_region(&screens[i], 0, 0,
+				    DisplayWidth(display, i),
+				    DisplayHeight(display, i)); 
+			else
+				new_region(&screens[i],
+				    ci->x, ci->y, ci->width, ci->height);
+		}
+		if (ci)
+			XRRFreeCrtcInfo(ci);
+		XRRFreeScreenResources(sr);
+	} else
+#endif /* SWM_XRR_HAS_CRTC */
+	{
+		new_region(&screens[i], 0, 0, DisplayWidth(display, i),
+		    DisplayHeight(display, i)); 
+	}
+}
+
+void
+screenchange(XEvent *e) {
+	XRRScreenChangeNotifyEvent	*xe = (XRRScreenChangeNotifyEvent *)e;
+	struct swm_region		*r;
+	struct ws_win			*win;
+	int				i;
+
+	DNPRINTF(SWM_D_EVENT, "screenchange: %lu\n", xe->root);
+
+	if (!XRRUpdateConfiguration(e))
+		return;
+
+	/* silly event doesn't include the screen index */
+	for (i = 0; i < ScreenCount(display); i++)
+		if (screens[i].root == xe->root)
+			break;
+	if (i >= ScreenCount(display))
+		errx(1, "screenchange: screen not found\n");
+
+	/* brute force for now, just re-enumerate the regions */
+	scan_xrandr(i);
+
+	/* hide any windows that went away */
+	TAILQ_FOREACH(r, &screens[i].rl, entry)
+		TAILQ_FOREACH(win, &r->ws->winlist, entry)
+			XUnmapWindow(display, win->id);
+	stack();
+}
+
+void
+setup_screens(void)
+{
 	Window			d1, d2, *wins = NULL;
 	XWindowAttributes	wa;
-	struct swm_region	*r;
 	unsigned int		no;
-	int			errorbase, major, minor;
-	int			ncrtc = 0, w = 0;
         int			i, j, k;
+	int			errorbase, major, minor;
 	struct workspace	*ws;
 	int			ws_idx_atom;
 
@@ -2402,13 +2647,20 @@ setup_screens(void)
 		errx(1, "calloc: screens");
 
 	ws_idx_atom = XInternAtom(display, "_SWM_WS", False);
-	
+
+	/* initial Xrandr setup */
+	xrandr_support = XRRQueryExtension(display,
+	    &xrandr_eventbase, &errorbase);
+	if (xrandr_support)
+		if (XRRQueryVersion(display, &major, &minor) && major < 1)
+       	               	xrandr_support = 0;
 
 	/* map physical screens */
 	for (i = 0; i < ScreenCount(display); i++) {
 		DNPRINTF(SWM_D_WS, "setup_screens: init screen %d\n", i);
 		screens[i].idx = i;
 		TAILQ_INIT(&screens[i].rl);
+		TAILQ_INIT(&screens[i].orl);
 		screens[i].root = RootWindow(display, i);
 
 		/* set default colors */
@@ -2419,6 +2671,7 @@ setup_screens(void)
 		setscreencolor("rgb:a0/a0/a0", i + 1, SWM_S_COLOR_BAR_FONT);
 
 		/* init all workspaces */
+		/* XXX these should be dynamically allocated too */
 		for (j = 0; j < SWM_WS_MAX; j++) {
 			ws = &screens[i].ws[j];
 			ws->idx = j;
@@ -2433,83 +2686,38 @@ setup_screens(void)
 					    SWM_ARG_ID_STACKINIT);
 			ws->cur_layout = &layouts[0];
 		}
-
-		/* map virtual screens onto physical screens */
-		screens[i].xrandr_support = XRRQueryExtension(display,
-		    &xrandr_eventbase, &errorbase);
-		if (screens[i].xrandr_support)
-			if (XRRQueryVersion(display, &major, &minor) &&
-			    major < 1)
-                        	screens[i].xrandr_support = 0;
-
-#if 0	/* not ready for dynamic screen changes */
-		if (screens[i].xrandr_support)
-			XRRSelectInput(display,
-			    screens[r->s].root,
-			    RRScreenChangeNotifyMask);
-#endif
-
 		/* grab existing windows (before we build the bars)*/
 		if (!XQueryTree(display, screens[i].root, &d1, &d2, &wins, &no))
 			continue;
 
-#ifdef SWM_XRR_HAS_CRTC
-		sr = XRRGetScreenResources(display, screens[i].root);
-		if (sr == NULL)
-			new_region(&screens[i], &screens[i].ws[w],
-			    0, 0, DisplayWidth(display, i),
-			    DisplayHeight(display, i)); 
-		else 
-			ncrtc = sr->ncrtc;
+		scan_xrandr(i);
 
-		for (c = 0, ci = NULL; c < ncrtc; c++) {
-			ci = XRRGetCrtcInfo(display, sr, sr->crtcs[c]);
-			if (ci->noutput == 0)
-				continue;
-
-			if (ci != NULL && ci->mode == None)
-				new_region(&screens[i], &screens[i].ws[w], 0, 0,
-				    DisplayWidth(display, i),
-				    DisplayHeight(display, i)); 
-			else
-				new_region(&screens[i], &screens[i].ws[w],
-				    ci->x, ci->y, ci->width, ci->height);
-			w++;
-		}
-		if (ci)
-			XRRFreeCrtcInfo(ci);
-		XRRFreeScreenResources(sr);
-#else
-		new_region(&screens[i], &screens[i].ws[w], 0, 0,
-		    DisplayWidth(display, i),
-		    DisplayHeight(display, i)); 
-#endif /* SWM_XRR_HAS_CRTC */
+		if (xrandr_support)
+			XRRSelectInput(display, screens[i].root,
+			    RRScreenChangeNotifyMask);
 
 		/* attach windows to a region */
 		/* normal windows */
-		if ((r = TAILQ_FIRST(&screens[i].rl)) == NULL)
-			errx(1, "no regions on screen %d", i);
-
-		for (i = 0; i < no; i++) {
-                        XGetWindowAttributes(display, wins[i], &wa);
-			if (!XGetWindowAttributes(display, wins[i], &wa) ||
+		for (j = 0; j < no; j++) {
+                        XGetWindowAttributes(display, wins[j], &wa);
+			if (!XGetWindowAttributes(display, wins[j], &wa) ||
 			    wa.override_redirect ||
-			    XGetTransientForHint(display, wins[i], &d1))
+			    XGetTransientForHint(display, wins[j], &d1))
 				continue;
 
 			if (wa.map_state == IsViewable ||
-			    getstate(wins[i]) == NormalState)
-				manage_window(wins[i]);
+			    getstate(wins[j]) == NormalState)
+				manage_window(wins[j]);
 		}
 		/* transient windows */
-		for (i = 0; i < no; i++) {
-			if (!XGetWindowAttributes(display, wins[i], &wa))
+		for (j = 0; j < no; j++) {
+			if (!XGetWindowAttributes(display, wins[j], &wa))
 				continue;
 
-			if (XGetTransientForHint(display, wins[i], &d1) &&
-			    (wa.map_state == IsViewable || getstate(wins[i]) ==
+			if (XGetTransientForHint(display, wins[j], &d1) &&
+			    (wa.map_state == IsViewable || getstate(wins[j]) ==
 			    NormalState))
-				manage_window(wins[i]);
+				manage_window(wins[j]);
                 }
                 if (wins) {
                         XFree(wins);
@@ -2580,10 +2788,27 @@ main(int argc, char *argv[])
 			bar_alarm = 0;
 			bar_update();
 		}
-		while(XPending(display)) {
+		while (XPending(display)) {
 			XNextEvent(display, &e);
-			if (handler[e.type])
-				handler[e.type](&e);
+			if (e.type < LASTEvent) {
+				if (handler[e.type])
+					handler[e.type](&e);
+				else
+					DNPRINTF(SWM_D_EVENT,
+					    "win: %lu unknown event: %d\n",
+					    e.xany.window, e.type);
+			} else {
+				switch (e.type - xrandr_eventbase) {
+				case RRScreenChangeNotify:
+					screenchange(&e);
+					break;
+				default:
+					DNPRINTF(SWM_D_EVENT,
+					    "win: %lu unknown xrandr event: "
+					    "%d\n", e.xany.window, e.type);
+					break;
+				}
+			}
 		}
 	}
 
